// This file is part of e93.
//
// e93 is free software; you can redistribute it and/or modify
// it under the terms of the e93 LICENSE AGREEMENT.
//
// e93 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// e93 LICENSE AGREEMENT for more details.
//
// You should have received a copy of the e93 LICENSE AGREEMENT
// along with e93; see the file "LICENSE.TXT".


// global definitions and structures known by the editor

#define	VERSION	"1.3"
#define	EDITION	"r4"

#define		Max(a,b) 	((a)>(b)?(a):(b))
#define		Min(a,b) 	((a)<(b)?(a):(b))

typedef	int					INT32;				// 32 bits signed
typedef	unsigned int		UINT32;				// 32 bits unsigned
typedef	short int			INT16;				// 16 bits signed
typedef	unsigned short int	UINT16;				// 16 bits unsigned
typedef	signed char			INT8;				// 8 bits signed
typedef	unsigned char		UINT8;				// 8 bits unsigned

// defines needed to handle text

#define	CHUNK_SIZE					1024		// size in bytes of data chunks used while editing text

#define	MAX_PATH_NAME_LENGTH		4096		// size in bytes maximum path name

#define	SELECTION_CHUNK_ELEMENTS	256			// number of SELECTIONELEMENTs in an array chunk
#define	UNDO_CHUNK_ELEMENTS			256			// number of undo elements in an undo chunk

#define	VIEW_MAX_STYLES				128			// maximum number of styles in a view (must be a power of 2)

#define	HORIZONTAL_SCROLL_THRESHOLD	32			// number of "pixels" to scroll horizontally whenever we do so

// modifier bits (can come from keyboard, or elsewhere (like mouse))

// masks
enum
{
	EEM_CAPSLOCK=0x00000001,
	EEM_SHIFT=0x00000002,
	EEM_CTL=0x00000004,
	EEM_MOD0=0x00000008,
	EEM_MOD1=0x00000010,
	EEM_MOD2=0x00000020,
	EEM_MOD3=0x00000040,
	EEM_MOD4=0x00000080,
	EEM_MOD5=0x00000100,
	EEM_MOD6=0x00000200,
	EEM_MOD7=0x00000400,
	EEM_STATE0=0x00F0000,		// traditionally used to keep track of double, triple, quadruple, etc... clicks of mouse
	EEM_STATE1=0x0F00000,
	EEM_STATE2=0xF000000
};

// shifts
enum
{
	EES_CAPSLOCK=0,
	EES_SHIFT=1,
	EES_CTL=2,
	EES_MOD0=3,
	EES_MOD1=4,
	EES_MOD2=5,
	EES_MOD3=6,
	EES_MOD4=7,
	EES_MOD5=8,
	EES_MOD6=9,
	EES_MOD7=10,
	EES_STATE0=16,
	EES_STATE1=20,
	EES_STATE2=24
};

// view event types

enum
{
	VET_KEYDOWN,			// saw key/virtual key press within view
	VET_KEYREPEAT,			// saw repeat within view
	VET_POSITIONVERTICAL,	// attempt to position the view to given absolute line
	VET_POSITIONHORIZONTAL,	// attempt to position the view to given absolute pixel
	VET_CLICK,				// initial click in view
	VET_CLICKHOLD			// send while click is being held in view
};

// virtual view keys

enum
{
	VVK_SCROLLUP,
	VVK_SCROLLDOWN,
	VVK_DOCUMENTPAGEUP,
	VVK_DOCUMENTPAGEDOWN,
	VVK_SCROLLLEFT,
	VVK_SCROLLRIGHT,
	VVK_DOCUMENTPAGELEFT,
	VVK_DOCUMENTPAGERIGHT,
// keys that are normally found on a keyboard, but not clearly associated with an ASCII code
// these are allowed to have modifier data sent with them
	VVK_LEFTARROW,
	VVK_RIGHTARROW,
	VVK_UPARROW,
	VVK_DOWNARROW,
	VVK_HOME,
	VVK_END,
	VVK_PAGEUP,
	VVK_PAGEDOWN,
	VVK_INSERT,
	VVK_BACKSPACE,
	VVK_DELETE,
	VVK_HELP,
	VVK_UNDO,
	VVK_REDO,
	VVK_UNDOTOGGLE,
	VVK_CUT,
	VVK_COPY,
	VVK_PASTE,
	VVK_FIND,
	VVK_RETURN,
	VVK_TAB,
	VVK_ESC,
	VVK_F1,
	VVK_F2,
	VVK_F3,
	VVK_F4,
	VVK_F5,
	VVK_F6,
	VVK_F7,
	VVK_F8,
	VVK_F9,
	VVK_F10,
	VVK_F11,
	VVK_F12,
	VVK_F13,
	VVK_F14,
	VVK_F15,
	VVK_F16,
	VVK_F17,
	VVK_F18,
	VVK_F19,
	VVK_F20
};

// menu relationships
enum
{
	MR_NEXTSIBLING,			// next sibling
	MR_PREVIOUSSIBLING,		// previous sibling
	MR_FIRSTCHILD,			// first child
	MR_LASTCHILD			// last child
};

// ways to home view (this type is used horizontally and vertically)
enum
{
	HT_NONE,				// do not move
	HT_LENIENT,				// if already on view, do not move, if off view, move as little as possible to get on view
	HT_SEMISTRICT,			// if already on view, do not move, if off view, move strict
	HT_STRICT				// move strict
};

// search types
enum
{
	ST_FIND,
	ST_FINDALL,
	ST_REPLACE,
	ST_REPLACEALL
};

typedef	struct editorFile EDITOR_FILE;		// this type is hidden from the editor (the editor only needs pointers to it)
typedef struct editorMenu EDITOR_MENU;		// menu definition (actual structure hidden from the editor)
typedef struct editorTask EDITOR_TASK;		// task definition (actual structure hidden)
typedef struct editorFont EDITOR_FONT;		// font definition (actual structure hidden)
typedef struct guiViewInfo GUI_VIEW_INFO;	// GUI's local view information

typedef struct tokenlist				// token list definition
{
	char	*token;
	int		tokenNum;
} TOKEN_LIST;

typedef struct variableBinding VARIABLE_BINDING;	// opaque variable binding

typedef struct variableTable
{
	VARIABLE_BINDING
		*variableBindingTable[256];		// array of heads of linked lists of variable bindings (crude hash table to get some speed when locating a binding)
	VARIABLE_BINDING
		*variableBindingListHead;		// head of list than runs through all variables
} VARIABLE_TABLE;

typedef struct
{
	INT32
		x,y;
	UINT32
		w,h;
} EDITOR_RECT;

typedef UINT32	EDITOR_COLOR;		// intensity values for each of the colors in RGB space (3 bytes R,G,B)

typedef struct chunkHeader
{
	struct chunkHeader
		*nextHeader,
		*previousHeader;
	UINT8
		*data;						// pointer to the data buffer for this chunk
	UINT32
		totalBytes;					// number of bytes of this chunk that are in use
	UINT32
		totalLines;					// number of new line characters in this chunk
} CHUNK_HEADER;

typedef struct
{
	CHUNK_HEADER
		*firstChunkHeader,			// pointer to the first chunk (NULL if there is none)
		*lastChunkHeader,			// pointer to the last chunk (NULL if there is none)
		*cacheChunkHeader;			// pointer to the cached chunk (NULL if there is none)
	UINT32
		totalBytes;					// total number of bytes of text
	UINT32
		totalLines;					// total number of new-line characters in text
	UINT32
		cacheBytes;					// total number of bytes up until the cacheChunkHeader
	UINT32
		cacheLines;					// total number of lines up until the cacheChunkHeader
} TEXT_UNIVERSE;

typedef struct arrayChunkHeader
{
	struct arrayChunkHeader
		*nextHeader,
		*previousHeader;
	UINT8
		*data;						// pointer to the element data for this chunk
	UINT32
		totalElements;				// number of elements of this chunk that are in use
} ARRAY_CHUNK_HEADER;

typedef struct
{
	ARRAY_CHUNK_HEADER
		*firstChunkHeader,			// points to the first header of the chunk array (or NULL if there are no chunks)
		*lastChunkHeader;
	UINT32
		totalElements;				// total number of elements in this universe
	UINT32
		maxElements,				// maximum number of elements in a chunk of this universe
		elementSize;				// size of an element of a chunk in this universe
} ARRAY_UNIVERSE;

typedef struct selectionUniverse SELECTION_UNIVERSE;
typedef struct styleUniverse STYLE_UNIVERSE;
typedef struct syntaxInstance SYNTAX_INSTANCE;

typedef struct undoElement
{
	UINT32
		frameNumber,				// tells which frame of undo this belongs to
		textStartPosition,			// position in text where replacement begins
		textLength,					// number of bytes to replace
		undoStartPosition,			// position in undo buffer of replacement text/style
		undoLength;					// length of replacement text
} UNDO_ELEMENT;

typedef struct
{
	ARRAY_UNIVERSE
		undoArray;					// array which is used to keep track of the individual undos
	TEXT_UNIVERSE
		*undoText;					// text used while undoing
	STYLE_UNIVERSE
		*undoStyles;				// styles used while undoing
} UNDO_GROUP;

typedef struct
{
	UINT32
		currentFrame;				// keeps track of the current undo frame index
	UINT32
		groupingDepth;				// knows the depth of begin/end undo group calls
	UINT32
		cleanElements;				// tells the number of elements in the undo-array when the buffer is considered "clean"
	bool
		cleanKnown;					// if false, cleanElements is meaningless and buffer is considered dirty
	bool
		undoActive;					// set true while undoing/redoing (tells not to delete redo information)
	bool
		undoing;					// set true while undoing (tells to place undo info into redo buffer)
	UNDO_GROUP
		undoGroup;					// used to keep track of undos
	UNDO_GROUP
		redoGroup;					// used to keep track of redos
} UNDO_UNIVERSE;

typedef struct markVisibility
{
	struct editorMark				// mark that is visible in the view
		*parentMark;
	struct editorView				// view that the mark is visible in
		*parentView;
	struct markVisibility			// links through list of marks visible on a given view
		*nextMark,
		*prevMark;
	struct markVisibility			// links through list of views a given mark is visible in
		*nextView,
		*prevView;
	EDITOR_COLOR
		backgroundColor;			// color to use when drawing background of visible mark
	SELECTION_UNIVERSE
		*heldMark;					// keeps track of the part of the mark in the view when mark is changed
} MARK_VISIBILITY;

typedef struct editorMark			// marks hanging off of editor universe
{
	char
		*markName;					// if NULL, then this mark is considered temporary
	SELECTION_UNIVERSE
		*selectionUniverse;			// the selection universe that defines this mark
	MARK_VISIBILITY
		*headVisibility,			// linked list of views this mark is visible in
		*tailVisibility;
	struct editorMark
		*nextMark;					// next mark in the list, NULL if no more
} EDITOR_MARK;

typedef struct EDITOR_BUFFER_HANDLE
{
	struct editorBuffer
		*theBuffer;					// points to the buffer being held
	EDITOR_BUFFER_HANDLE
		*nextHandle; 				// next in linked list of handles
};

typedef struct editorBuffer
{
	EDITOR_BUFFER_HANDLE
		*bufferHandles;				// when something needs to know when a buffer has been deleted, it can get a handle to it which the buffer code will clear on deletion
	bool
		shellBusy;					// used by the shell to make sure it cannot modify a buffer while it is in use by some previous shell command
	bool
		fromFile;					// tells if contents of this buffer represent an actual file
	char
		*contentName;				// name that describes the contents of the buffer (if fromFile=true, then this is full absolute pathname of file)
	TEXT_UNIVERSE
		*textUniverse;				// all the text
	UNDO_UNIVERSE
		*undoUniverse;				// undo/redo information
	SELECTION_UNIVERSE
		*selectionUniverse;			// selection and cursor information
	SELECTION_UNIVERSE
		*auxSelectionUniverse;		// aux selection and cursor information (allows tasks to write to universe at place other than cursor)
	EDITOR_MARK
		*theMarks;					// pointer to head of list of marks for this universe, NULL if none
	STYLE_UNIVERSE
		*styleUniverse;				// maintains style array
	SYNTAX_INSTANCE
		*syntaxInstance;			// keeps track of syntax highlighting (no syntax highlights if NULL)
	EDITOR_TASK
		*theTask;					// pointer to the task record for this buffer, or NULL if no task linked to this buffer
	UINT32
		taskBytes;					// how many bytes of data the last task added to theBuffer
	struct editorWindow
		*theWindow;					// pointer to document window for this buffer (if there is one, NULL if none)
	struct editorView
		*firstView;					// head of linked list of views onto this buffer
	VARIABLE_TABLE
		variableTable;				// table of variables linked to this buffer
	// NOTE: the members below help keep track of various things during replacements
	bool
		replaceHaveRange;			// tells if replacement range encountered yet
	UINT32
		replaceStartOffset,			// composite start offset in text
		replaceEndOffset;			// composite end offset in text (during the change)
	INT32
		replaceNumBytes;			// number of bytes added/(subtracted) from text
	// NOTE: the members below help keep track of various things during selection (this is a bit of a mess and should be simplified)
	UINT32
		anchorStartPosition,		// hold the ends of the selection while tracking
		anchorEndPosition;
	UINT32
		anchorLine;					// holds one corner of columnar selection while tracking
	INT32
		anchorX;
	UINT32
		columnarTopLine;			// when building a columnar selection, these remember the "rect" that bounds the selection
	INT32
		columnarLeftX;				// they are used when adding to the selection
	UINT32
		columnarBottomLine;
	INT32
		columnarRightX;
	bool
		haveStartX,					// tell if desiredStartX/desiredEndX are valid
		haveEndX;
	INT32
		desiredStartX,				// desired X position of the start of the selection when expanding/reducing, or of the cursor when no selection (used when moving the cursor up or down)
		desiredEndX;
	bool
		haveCurrentEnd;				// true if there is a current end of the selection being moved (used when extending the selection with the cursor keys)
	INT32
		currentIsStart;				// true if the current end being moved is the start
	struct editorBuffer
		*nextBuffer;				// next buffer in linked list of buffers
} EDITOR_BUFFER;

typedef struct editorWindow
{
	UINT16
		windowType;					// which type of editor window this is
	EDITOR_BUFFER
		*theBuffer;					// if this is a document window, then this points to the buffer, otherwise it is NULL
	void
		*windowInfo;				// pointer to gui specific structure that keeps all info needed for this window type
	void
		*userData;					// pointer to any additional gui specific info for this window (like the gui's actual window structure)
} EDITOR_WINDOW;

typedef struct viewDescriptor		// describes a view to creation routines
{
	EDITOR_BUFFER
		*theBuffer;					// the buffer on which to view
	EDITOR_RECT
		theRect;					// position and size within parent window
	UINT32
		topLine;
	INT32
		leftPixel;
	UINT32
		tabSize;
	EDITOR_COLOR
		backgroundColor;			// default color to use when drawing background (style 0)
	EDITOR_COLOR
		foregroundColor;			// default color to use when drawing foreground (style 0)
	char
		*fontName;					// default font to use (style 0)
	bool
		active;						// tells if the view is created active or not
	void
		(*viewTextChangedVector)(struct editorView *theView);	// this vector is called by the view when it has finished a text change
	void
		(*viewSelectionChangedVector)(struct editorView *theView);	// this vector is called by the view when it has finished a selection change
	void
		(*viewPositionChangedVector)(struct editorView *theView);	// this vector is called by the view when it has finished a position change
} EDITOR_VIEW_DESCRIPTOR;

typedef struct
{
	EDITOR_FONT
		*theFont;					// pointer to font in use (NULL if default should be used)
	bool
		haveForegroundColor,		// tells if foreground color has been specified
		haveBackgroundColor;		// tells if background color has been specified
	EDITOR_COLOR
		foregroundColor,			// color to use when drawing foreground
		backgroundColor;			// color to use when drawing background
} VIEW_STYLE;

// NOTE: it is possible to have multiple views of the same editor universe
typedef struct editorView
{
	EDITOR_BUFFER
		*parentBuffer;				// the buffer which this view is placed
	EDITOR_WINDOW
		*parentWindow;				// the window that contains this view
	struct editorView
		*nextBufferView;			// next view of the buffer that this view is on
	bool
		active;						// if true, then view is displayed as active
	bool
		wantHome;					// used during buffer update, tells if given view should be homed
	UINT32
		tabSize;					// number of columns per tab
	bool
		haveSelectionForegroundColor,	// tells if color has been specified
		haveSelectionBackgroundColor;	// tells if color has been specified
	EDITOR_COLOR
		selectionForegroundColor,	// color to use when drawing selection foreground
		selectionBackgroundColor;	// color to use when drawing selection background
	MARK_VISIBILITY
		*headVisibility,			// linked list of marks visible in this view
		*tailVisibility;
	void
		(*viewTextChangedVector)(struct editorView *theView);	// this vector is called by the view when it has finished a text change
	void
		(*viewSelectionChangedVector)(struct editorView *theView);	// this vector is called by the view when it has finished a selection change
	void
		(*viewPositionChangedVector)(struct editorView *theView);	// this vector is called by the view when it has finished a position change
	// all of these variables are used to facilitate updates of views
	INT32
		updateStartX;				// used during update, keeps the X position into the line being updated (if it is on the view)
	UINT32
		startPosition,				// used during update, keep track of the character positions that the view spans
		endPosition;
	VIEW_STYLE
		viewStyles[VIEW_MAX_STYLES];	// array of styles
	SELECTION_UNIVERSE
		*heldSelection;				// keeps track of selections that were in the view across calls to Start/EndSelectionChange
	STYLE_UNIVERSE
		*heldStyle;					// keeps track of styles that were in the view across calls to Start/EndStyleChange
	GUI_VIEW_INFO
		*viewInfo;					// pointer to gui specific info about this view (gui needs to remember font, size, bounds, top line, etc)
} EDITOR_VIEW;

typedef struct
{
	bool
		isVirtual;					// tells if the keyCode is a virtual one or not
	UINT32
		keyCode;					// gives keyboard position of non-modifier key that was hit, or the virtual key number if it was a virtual key
	UINT32
		modifiers;					// modifier bits (enumerated above)
} EDITOR_KEY;

typedef struct
{
	UINT32
		modifiers;					// keyboard style modifier bits (enumerated above)
	UINT32
		position;					// absolute position (in pixels, or lines)
} VIEW_POS_EVENT_DATA;

typedef struct
{
	UINT32
		keyCode;					// mouse button (0-n)
	UINT32
		modifiers;					// keyboard style modifier bits (enumerated above)
	INT32
		xClick,
		yClick;						// view relative click point
} VIEW_CLICK_EVENT_DATA;

typedef struct						// high level view events (sent to the shell from the view)
{
	UINT16
		eventType;
	EDITOR_VIEW
		*theView;					// the view that this event belongs to
	void
		*eventData;					// pointer to event specific data for event
} VIEW_EVENT;

