# This file takes the place of a proper XML parser and a XSL sytlesheet
# the input XML must contain TITLE and DESCRIPTION

# tagTable assosiates the tag name with the section name that will appear in the html
# TODO: put all the tag names into TclVars and replace all the literals in this file

# required
	set	tagTable(TITLE)				"TITLE"
	set	tagTable(DESCRIPTION)		"DESCRIPTION"

# the SYNOPSIS section is generated

#defaults
	# defaults to "Built-in" if not specified
	set	tagTable(TYPE)				"TYPE"

	# defaults to fist sentance of DESCRIPTION if not specified
	set	tagTable(SUMMARY)			"SUMMARY"

	# default to "None" if not specified
	set	tagTable(INPUT)				"INPUT"
	set	tagTable(OUTPUT)			"OUTPUT"

	# defaults to "TITLE always returns a status code of 0." if not specified
	set	tagTable(STATUS)			"STATUS"

	# 0-N
	set	tagTable(PARAMETERS)		"PARAMETER"
	set	tagTable(OPTION)			"OPTION"

	# 0-N PARAMETER tags (it should contain a PARAMDESCRIPTION for each one it includes)
	set	tagTable(PARAMDESCRIPTION)	"PARAMDESCRIPTION"

	# 0-N OPTION tags (it should contain an OPTDESCRIPTION for each one it includes)
	set	tagTable(OPTDESCRIPTION)	"OPTDESCRIPTION"

#optional
	# these sections will not appear unless explcitly included
	set	tagTable(EXAMPLES)			"EXAMPLES"
	set	tagTable(LIMITATIONS)		"LIMITATIONS"
	set	tagTable(SEEALSO)			"SEE ALSO"
	set	tagTable(LOCATION)			"LOCATION"

proc makeSection {heading contents} \
{
	return "<H3>\n$heading\n</H3>\n<BLOCKQUOTE>\n<P>$contents</P>\n</BLOCKQUOTE>"
}

proc makeHTMLCommands {srcFile destfile} \
{
	global tagTable

	# load the source file
	set inChan [open $srcFile]
	set body [read $inChan]
	close $inChan

	set title [getTaggedText "TITLE" $body]
	set summary [getTaggedText "SUMMARY" $body]
	if {$summary == ""} \
	{
		set summary [getTaggedText "DESCRIPTION" $body]
		set exp \[^\.<]*
		regexp -nocase $exp $summary summary
	}

	set type [getTaggedText "TYPE" $body]
	if {$type == ""} \
	{
		set type "Built-in"
	}

	# open a destination file
	set outfile [open $destfile w+]
	fconfigure $outfile -translation lf

	puts $outfile "<TITLE>$title</TITLE>"
	puts $outfile "<LINK REL=\"STYLESHEET\" HREF=\"../cmdref.css\" CHARSET=\"ISO-8859-1\" TYPE=\"text/css\">"
	puts $outfile "<!-- <OBJECT type=\"text/summary\">\n$summary\n</OBJECT> -->"
	puts $outfile "<TABLE WIDTH=\"750\" BORDER=\"0\" CELLSPACING=\"0\" CELLPADDING=\"0\">\n<TR>\n<TD WIDTH=\"500\" ALIGN=\"LEFT\"><H1>$title</H1></TD>\n<TD WIDTH=\"250\" ALIGN=\"RIGHT\"><H1>$type</H1></TD>\n</TR>\n</TABLE>"

	# manditory sections

	# build SYNOPSIS section
	set section SYNOPSIS
	set contents "<CODE>$title<VAR>"
	set options [getAllTaggedText OPTION $body]
	foreach option $options \
	{
		append contents " [string trim  $option]"
	}
	set parameters [getAllTaggedText PARAMETER $body]
	foreach parameter $parameters \
	{
		append contents " [string trim  $parameter]"
	}
	append contents "</VAR></CODE>"
	puts $outfile [makeSection "$section" $contents]

	# get manditory sections, use default if they are not part of the XML
	foreach section {DESCRIPTION INPUT OUTPUT} \
	{
		set contents [getTaggedText "$section" $body]
		if {$contents == ""} \
		{
			set contents "None"
		}
		
		puts $outfile [makeSection "$section" $contents]
	}

	# build STATUS section
	set section "STATUS"
	set contents [getTaggedText $section $body]
	if {$contents == ""} \
	{
		set contents "$title always returns a status code of 0."
	}
	puts $outfile [makeSection "$section" $contents]

	# build PARAMETERS section
	set section PARAMETERS
	set contents ""
	if {[llength $parameters]==0} \
	{
		set contents "None"
	}
	foreach parameter $parameters \
	{
		set parameter [string trim $parameter]
		append contents "<VAR>$parameter</VAR>\n<BLOCKQUOTE>\n<P>[getTaggedText "PARAMDESCRIPTION" $body " name=\"$parameter\""]</P>\n</BLOCKQUOTE>"
	}
	puts $outfile [makeSection "$section" $contents]


	# build OPTIONS section
	set section OPTIONS
	set contents ""
	if {[llength $options]==0} \
	{
		set contents "None"
	}
	foreach option $options \
	{
		set option [string trim $option]
		append contents "<VAR>$option</VAR>\n<BLOCKQUOTE>\n<P>[getTaggedText "OPTDESCRIPTION" $body " name=\"$option\""]</P>\n</BLOCKQUOTE>"
	}
	puts $outfile [makeSection "$section" $contents]


	# optional sections
	foreach section {EXAMPLES LIMITATIONS SEEALSO LOCATION} \
	{
		set contents [getTaggedText "$section" $body]
		if {$contents != ""} \
		{
			puts $outfile [makeSection $tagTable($section) $contents]
		}
	}

	close $outfile
}

proc makeHTMLChapters {srcFile destfile} \
{
	# load the source file
	set inChan [open $srcFile]
	set body [read $inChan]
	close $inChan

	set title [getTaggedText "TITLE" $body]

	set summary [getTaggedText "SUMMARY" $body]
	if {$summary == ""} \
	{
		set summary [getTaggedText "BODY" $body]
		set exp \[^\.<]*
		regexp -nocase $exp $summary summary
	}

	# open a destination file
	set outfile [open $destfile w+]
	fconfigure $outfile -translation lf

	puts $outfile "<TITLE>$title</TITLE>"
	puts $outfile "<LINK REL=\"STYLESHEET\" HREF=\"cmdref.css\" CHARSET=\"ISO-8859-1\" TYPE=\"text/css\">"
	puts $outfile "<!-- <OBJECT type=\"text/summary\">\n$summary\n</OBJECT> -->"
	puts $outfile "<CENTER>\n<H1>\n$title\n</H1>\n</CENTER>"
	# if we wanted sections we would do it here
	puts $outfile "<P>[getTaggedText "BODY" $body]</P>"
	
	close $outfile
}

set xmlDir xml
set htmlDir html

# remove the old html dir
file delete -force $htmlDir

# make the html dir
file mkdir $htmlDir/commands

# make the commands
set sourceFiles [lsort [glob -nocomplain $xmlDir/commands/*.xml]]
foreach srcFile $sourceFiles \
{
	set destfile $htmlDir/commands/[file tail [file rootname $srcFile]].html
	makeHTMLCommands $srcFile $destfile
}

# make the chapters
set sourceFiles [lsort [glob -nocomplain $xmlDir/*.xml]]
foreach srcFile $sourceFiles \
{
	set destfile $htmlDir/[file tail [file rootname $srcFile]].html
	makeHTMLChapters $srcFile $destfile
}

# copy any cascading style sheets
catch {file copy -force [glob $xmlDir/*.css] $htmlDir/}

# copy images, if any
catch {file copy -force [glob $xmlDir/*.jpg]  $htmlDir}
catch {file copy -force [glob $xmlDir/*.gif]  $htmlDir}
catch {file copy -force [glob $xmlDir/commands/*.jpg]  $htmlDir/commands/}
catch {file copy -force [glob $xmlDir/commands/*.gif]  $htmlDir/commands/}

