// This file is part of e93.
//
// e93 is free software; you can redistribute it and/or modify
// it under the terms of the e93 LICENSE AGREEMENT.
//
// e93 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// e93 LICENSE AGREEMENT for more details.
//
// You should have received a copy of the e93 LICENSE AGREEMENT
// along with e93; see the file "LICENSE.TXT".


// This include file is used to communicate with the GUI
// it contains all prototypes for porting functions that editor
// expects from the GUI implementation.


// ALL the routines that the editor requires from the GUI ------------------------------------------------------------------------------------------

void EditorBeep();
void ReportMessage(char *format,...);
void GetEditorScreenDimensions(UINT32 *theWidth,UINT32 *theHeight);

void GetEditorViewGraphicToTextPosition(EDITOR_VIEW *theView,UINT32 linePosition,INT32 xPosition,UINT32 *betweenOffset,UINT32 *charOffset);
void GetEditorViewTextToGraphicPosition(EDITOR_VIEW *theView,UINT32 thePosition,INT32 *xPosition,bool limitMax,UINT32 *slopLeft,UINT32 *slopRight);
void GetEditorViewTextInfo(EDITOR_VIEW *theView,UINT32 *topLine,UINT32 *numLines,INT32 *leftPixel,UINT32 *numPixels);
void InvalidateViewPortion(EDITOR_VIEW *theView,UINT32 startLine,UINT32 endLine,UINT32 startPixel,UINT32 endPixel);
void InvalidateViewSelection(EDITOR_VIEW *theView,SELECTION_UNIVERSE *theSelection);
void InvalidateView(EDITOR_VIEW *theView);
void ScrollViewPortion(EDITOR_VIEW *theView,UINT32 startLine,UINT32 endLine,INT32 numLines,UINT32 startPixel,UINT32 endPixel,INT32 numPixels);
void SetViewTopLeft(EDITOR_VIEW *theView,UINT32 newTopLine,INT32 newLeftPixel);
void SetEditorViewTopLine(EDITOR_VIEW *theView,UINT32 lineNumber);
void ResetEditorViewCursorBlink(EDITOR_VIEW *theView);
void ViewStyleFontsChanged(EDITOR_VIEW *theView);
void ViewEndChange(EDITOR_VIEW *theView);
void ViewStartChange(EDITOR_VIEW *theView);
void DisposeViewInfo(EDITOR_VIEW *theView);
bool CreateViewInfo(EDITOR_VIEW *theView,EDITOR_VIEW_DESCRIPTOR *theDescriptor);

bool EditorColorNameToColor(char *colorName,EDITOR_COLOR *theColor);

void UpdateEditorWindows();

void MinimizeDocumentWindow(EDITOR_WINDOW *theWindow);
void UnminimizeDocumentWindow(EDITOR_WINDOW *theWindow);
void SetTopDocumentWindow(EDITOR_WINDOW *theWindow);
EDITOR_WINDOW *GetActiveDocumentWindow();
EDITOR_WINDOW *GetTopDocumentWindow();
bool GetSortedDocumentWindowList(UINT32 *numElements,EDITOR_WINDOW ***theList);
EDITOR_VIEW *GetDocumentWindowCurrentView(EDITOR_WINDOW *theWindow);

bool GetDocumentWindowTitle(EDITOR_WINDOW *theWindow,char *theTitle,UINT32 stringBytes);
bool SetDocumentWindowTitle(EDITOR_WINDOW *theWindow,char *theTitle);
bool GetEditorDocumentWindowForegroundColor(EDITOR_WINDOW *theWindow,EDITOR_COLOR *foregroundColor);
bool SetEditorDocumentWindowForegroundColor(EDITOR_WINDOW *theWindow,EDITOR_COLOR foregroundColor);
bool GetEditorDocumentWindowBackgroundColor(EDITOR_WINDOW *theWindow,EDITOR_COLOR *backgroundColor);
bool SetEditorDocumentWindowBackgroundColor(EDITOR_WINDOW *theWindow,EDITOR_COLOR backgroundColor);
bool GetEditorDocumentWindowRect(EDITOR_WINDOW *theWindow,EDITOR_RECT *theRect);
bool SetEditorDocumentWindowRect(EDITOR_WINDOW *theWindow,EDITOR_RECT *theRect);
EDITOR_WINDOW *OpenDocumentWindow(EDITOR_BUFFER *theBuffer,EDITOR_RECT *theRect,char *theTitle,char *fontName,UINT32 tabSize,EDITOR_COLOR foreground,EDITOR_COLOR background);
void CloseDocumentWindow(EDITOR_WINDOW *theWindow);

char *CreateWindowTitleFromPath(char *absolutePath);
char *CreateAbsolutePath(char *relativePath);
bool LocateStartupScript(char *scriptPath);

EDITOR_MENU *GetRelatedEditorMenu(EDITOR_MENU *theMenu,int menuRelationship);
void DeactivateEditorMenu(EDITOR_MENU *theMenu);
void ActivateEditorMenu(EDITOR_MENU *theMenu);
char *GetEditorMenuName(EDITOR_MENU *theMenu);
bool SetEditorMenuName(EDITOR_MENU *theMenu,const char *theName);
char *GetEditorMenuAttributes(EDITOR_MENU *theMenu);
bool SetEditorMenuAttributes(EDITOR_MENU *theMenu,const char *theAttributes);
char *GetEditorMenuDataText(EDITOR_MENU *theMenu);
bool SetEditorMenuDataText(EDITOR_MENU *theMenu,const char *theDataText);
bool GetEditorMenu(int argc,const char *argv[],EDITOR_MENU **theMenu);
void DisposeEditorMenu(EDITOR_MENU *theMenu);
EDITOR_MENU *CreateEditorMenu(int argc,const char *argv[],int menuRelationship,const char *theName,const char *theAttributes,const char *dataText,bool active);

bool EditorKeyNameToKeyCode(char *theName,UINT32 *theCode);
char *EditorKeyCodeToKeyName(UINT32 theCode);
bool EditorGetKeyPress(UINT32 *keyCode,UINT32 *editorModifiers,bool wait,bool clearBuffered);

void EditorQuit();
void EditorEventLoop(int argc,char *argv[]);
void ClearAbort();
bool CheckAbort();

void ShowBusy();
void ShowNotBusy();

bool ImportClipboard();
bool ExportClipboard();

void GetFontName(EDITOR_FONT *theFont,char *nameBuffer,UINT32 bufferLength);
void FreeFont(EDITOR_FONT *theFont);
EDITOR_FONT *LoadFont(const char *theName);

EDITOR_FILE *OpenEditorReadFile(char *thePath);
EDITOR_FILE *OpenEditorWriteFile(char *thePath);
bool ReadEditorFile(EDITOR_FILE *theFile,UINT8 *theBuffer,UINT32 bytesToRead,UINT32 *bytesRead);
bool WriteEditorFile(EDITOR_FILE *theFile,UINT8 *theBuffer,UINT32 bytesToWrite,UINT32 *bytesWritten);
void CloseEditorFile(EDITOR_FILE *theFile);

bool OkDialog(const char *theText);
bool OkCancelDialog(const char *theText,bool *cancel);
bool YesNoCancelDialog(const char *theText,bool *yes,bool *cancel);
bool GetSimpleTextDialog(const char *theTitle,char *enteredText,UINT32 stringBytes,bool *cancel);
bool SearchReplaceDialog(EDITOR_BUFFER *searchBuffer,EDITOR_BUFFER *replaceBuffer,bool *backwards,bool *wrapAround,bool *selectionExpr,bool *ignoreCase,bool *limitScope,bool *replaceProc,UINT16 *searchType,bool *cancel);
bool SimpleListBoxDialog(const char *theTitle,UINT32 numElements,const char **listElements,bool *selectedElements,bool *cancel);
bool OpenFileDialog(const char *theTitle,char *fullPath,UINT32 stringBytes,char ***listElements,bool *cancel);
void FreeOpenFileDialogPaths(char **thePaths);
bool SaveFileDialog(const char *theTitle,char *fullPath,UINT32 stringBytes,bool *cancel);
bool ChoosePathDialog(const char *theTitle,char *fullPath,UINT32 stringBytes,bool *cancel);
bool ChooseFontDialog(const char *theTitle,char *theFont,UINT32 stringBytes,bool *cancel);

bool WriteBufferTaskData(EDITOR_BUFFER *theBuffer,UINT8 *theData,UINT32 numBytes);
bool SendEOFToBufferTask(EDITOR_BUFFER *theBuffer);
void DisconnectBufferTask(EDITOR_BUFFER *theBuffer);
bool UpdateBufferTask(EDITOR_BUFFER *theBuffer,bool *didUpdate);
bool KillBufferTask(EDITOR_BUFFER *theBuffer);

void *MNewPtr(UINT32 theSize);
void *MNewPtrClr(UINT32 theSize);
void MDisposePtr(void *thePointer);
void MMoveMem(void *source,void *dest,UINT32 numBytes);
UINT32 MGetNumAllocatedPointers();

char *GetEditorLocalVersion();
bool AddSupplementalShellCommands(Tcl_Interp *theInterpreter);
bool InitEnvironment();
void UnInitEnvironment();
bool EarlyInit();
void EarlyUnInit();

