// Font handling
// Copyright (C) 2000 Core Technologies.

// This file is part of e93.
//
// e93 is free software; you can redistribute it and/or modify
// it under the terms of the e93 LICENSE AGREEMENT.
//
// e93 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// e93 LICENSE AGREEMENT for more details.
//
// You should have received a copy of the e93 LICENSE AGREEMENT
// along with e93; see the file "LICENSE.TXT".

#include	"includes.h"

static unsigned int
	r_max,
	g_max,
	b_max,
	r_shift,
	g_shift,
	b_shift;

unsigned long EditorColorToXPixel(EDITOR_COLOR inColor)
// Convert an editor color into an X pixel color
{
	unsigned long
		pixel;
	unsigned int
		r,
		g,
		b;

	r=(inColor>>16)&0xFF;
	g=(inColor>>8)&0xFF;
	b=(inColor>>0)&0xFF;
	pixel=(((r_max+1)*r/256)<<r_shift)|(((g_max+1)*g/256)<<g_shift)|(((b_max+1)*b/256)<<b_shift);
	return(pixel);
}

bool EditorColorNameToColor(char *colorName,EDITOR_COLOR *theColor)
// Convert a machine specific color name into an editor color
// if the name does not map to a color, return false
// NOTE: it is reasonable for this routine to always return
// false on machines which do not assign names to colors
{
	XColor
		exactResult,
		screenResult;

	if(XLookupColor(xDisplay,DefaultColormap(xDisplay,xScreenNum),colorName,&exactResult,&screenResult))
	{
		*theColor=((screenResult.red>>8)<<16)+((screenResult.green>>8)<<8)+((screenResult.blue>>8));
		return(true);
	}
	return(false);
}

void UnInitColors()
// Undo what InitColors did
{
}

bool InitColors()
// Initialize editor color handling
// If there is a fatal problem, SetError, return false
{
	int
		displayDepth;
	XImage
		*tmpImage;
		
	displayDepth=DefaultDepth(xDisplay,xScreenNum);		// get the color bit depth
	if(displayDepth>=15)								// needs to 15 bits or better to run
	{
		if((tmpImage=XCreateImage(xDisplay,DefaultVisual(xDisplay,xScreenNum),displayDepth,ZPixmap,0,NULL,1,1,8,0)))
		{
			// Setup global color masks and shifts
			r_max=tmpImage->red_mask;
			g_max=tmpImage->green_mask;
			b_max=tmpImage->blue_mask;
			XDestroyImage(tmpImage);					// lose the image we created

			// normalize masks, work out shifts
			for(r_shift=0;!(r_max&1);r_shift++)
			{
				r_max>>=1;
			}
			for(g_shift=0;!(g_max&1);g_shift++)
			{
				g_max>>=1;
			}
			for(b_shift=0;!(b_max&1);b_shift++)
			{
				b_max>>=1;
			}
			return(true);
		}
	}
	else
	{
		SetError("color","BadBitDepth","This program requires a color bit depth of 15 bits or greater.");
	}
	return(false);
}
