// This file is part of e93.
//
// e93 is free software; you can redistribute it and/or modify
// it under the terms of the e93 LICENSE AGREEMENT.
//
// e93 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// e93 LICENSE AGREEMENT for more details.
//
// You should have received a copy of the e93 LICENSE AGREEMENT
// along with e93; see the file "LICENSE.TXT".


// defines and data structures for text boxes

typedef struct textBox
{
	EDITOR_WINDOW
		*parentWindow;					// the window that contains this text box
	EDITOR_RECT
		theRect;						// position and size within parent window
	EDITOR_COLOR
		focusBackgroundColor,			// color to use when drawing background when we have the focus
		focusForegroundColor,			// color to use when drawing foreground when we have the focus
		nofocusBackgroundColor,			// color to use when drawing background when we do not have focus
		nofocusForegroundColor;			// color to use when drawing foreground when we do not have focus
	EDITOR_VIEW
		*theView;						// the view that this text box item made
} TEXT_BOX;

typedef struct textBoxDescriptor		// describes a list box to creation routines
{
	EDITOR_BUFFER
		*theBuffer;						// the buffer on which to view
	EDITOR_RECT
		theRect;						// position and size within parent window
	UINT32
		topLine;
	INT32
		leftPixel;
	UINT32
		tabSize;
	EDITOR_COLOR
		focusBackgroundColor,			// color to use when drawing background when it has focus
		focusForegroundColor,			// color to use when drawing foreground when it has focus
		nofocusBackgroundColor,			// color to use when drawing background when it does not have focus
		nofocusForegroundColor;			// color to use when drawing foreground when it does not have focus
	char
		*fontName;						// name of font to use
} TEXT_BOX_DESCRIPTOR;

bool PointInTextBox(TEXT_BOX *theTextBox,INT32 x,INT32 y);
void DrawTextBox(TEXT_BOX *theTextBox,bool hasFocus);
bool TrackTextBox(TEXT_BOX *theTextBox,XEvent *theEvent);
void RepositionTextBox(TEXT_BOX *theTextBox,EDITOR_RECT *newRect);
void HandleTextBoxKeyEvent(TEXT_BOX *theTextBox,XEvent *theEvent);
void ActivateTextBox(TEXT_BOX *theTextBox);
void DeactivateTextBox(TEXT_BOX *theTextBox);
void HandleTextBoxPeriodicProc(TEXT_BOX *theTextBox);
TEXT_BOX *CreateTextBox(EDITOR_WINDOW *theWindow,TEXT_BOX_DESCRIPTOR *theDescription);
void DisposeTextBox(TEXT_BOX *theTextBox);
bool CreateTextBoxItem(DIALOG_ITEM *theItem,void *theDescriptor);
