// This file is part of e93.
//
// e93 is free software; you can redistribute it and/or modify
// it under the terms of the e93 LICENSE AGREEMENT.
//
// e93 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// e93 LICENSE AGREEMENT for more details.
//
// You should have received a copy of the e93 LICENSE AGREEMENT
// along with e93; see the file "LICENSE.TXT".


#define	VIEW_TOP_MARGIN			4		// number of pixels to leave at the top of the view
#define	VIEW_BOTTOM_MARGIN		4		// number of pixels to leave at the bottom of the view
#define	VIEW_LEFT_MARGIN		10		// number of pixels to the left of the text in the view that we call the left hand side

typedef struct guiViewInfo
{
	Window
		cursorWindow;							// x InputOnly window used to make the cursor change to caret within the view (this is ALL it is used for!)
	EDITOR_RECT
		bounds;									// boundary of this view within its parent window
	EDITOR_RECT
		textBounds;								// roughly the bounds, but rounded down for the current font
	UINT32
		topLine;								// current first line of edit universe that is displayed in the view
	INT32
		leftPixel;								// current offset from left where text is being drawn in this view
	UINT32
		lineHeight;								// number of pixels per line vertically
	UINT32
		lineAscent;								// tells where the baseline of text is for each line in the view
	UINT32
		maxLeftOverhang;						// gives the maximum font character left overhang for any character in any style of this view
	UINT32
		maxRightOverhang;						// gives the maximum font character right overhang for any character in any style of this view
	UINT32
		numLines;								// number of lines that can be displayed
	bool
		cursorOn;								// if true, and no selection, and view active, and blinkState, then cursor will be shown
	bool
		blinkState;								// tells what state (true = on) (false = off) the cursor blink is currently in
	UINT32
		nextBlinkTime;							// tells the next time that the cursor should change blink state
	bool
		oldCursorShowing;						// holds showing state of cursor across Start/EndCursorChange
} GUI_VIEW_INFO;

void HandleViewKeyEvent(EDITOR_VIEW *theView,XEvent *theEvent);
void GetEditorViewGraphicToTextPosition(EDITOR_VIEW *theView,UINT32 linePosition,INT32 xPosition,UINT32 *betweenOffset,UINT32 *charOffset);
void GetEditorViewTextToGraphicPosition(EDITOR_VIEW *theView,UINT32 thePosition,INT32 *xPosition,bool limitMax,UINT32 *slopLeft,UINT32 *slopRight);
void GetEditorViewTextInfo(EDITOR_VIEW *theView,UINT32 *topLine,UINT32 *numLines,INT32 *leftPixel,UINT32 *numPixels);
void DrawView(EDITOR_VIEW *theView);
void SetViewBounds(EDITOR_VIEW *theView,EDITOR_RECT *newBounds);
void InvalidateViewPortion(EDITOR_VIEW *theView,UINT32 startLine,UINT32 endLine,UINT32 startPixel,UINT32 endPixel);
void InvalidateViewSelection(EDITOR_VIEW *theView,SELECTION_UNIVERSE *theSelection);
void InvalidateView(EDITOR_VIEW *theView);
void ScrollViewPortion(EDITOR_VIEW *theView,UINT32 startLine,UINT32 endLine,INT32 numLines,UINT32 startPixel,UINT32 endPixel,INT32 numPixels);
void SetViewTopLeft(EDITOR_VIEW *theView,UINT32 newTopLine,INT32 newLeftPixel);
void SetEditorViewTopLine(EDITOR_VIEW *theView,UINT32 lineNumber);
void ResetEditorViewCursorBlink(EDITOR_VIEW *theView);
void ViewCursorTask(EDITOR_VIEW *theView);
bool TrackView(EDITOR_VIEW *theView,XEvent *theEvent);
void ViewStyleFontsChanged(EDITOR_VIEW *theView);
void ViewEndChange(EDITOR_VIEW *theView);
void ViewStartChange(EDITOR_VIEW *theView);
void DisposeViewInfo(EDITOR_VIEW *theView);
bool CreateViewInfo(EDITOR_VIEW *theView,EDITOR_VIEW_DESCRIPTOR *theDescriptor);
